#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec 23, 2018
email    : Nasy <nasyxx+python@gmail.com>
filename : tools.py
project  : nacf
license  : GPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
# Standard Library
from inspect import Parameter, Signature, signature

# Prelude
from nalude import flatten

# Local Packages
from .types import Url, Func, Optional


def format_url(url: Optional[Url]) -> Optional[Url]:
    """Add schema to url."""
    return (
        url if url and url.startswith("http") else url and ("https://" + url)
    )


def generate_sig(func: Func, *params: Parameter) -> Signature:
    """Generate signature with param from func."""
    return signature(func).replace(
        parameters=list(signature(func).parameters.values()) + list(params)
    )


__all__ = ["format_url", "flatten"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec 23, 2018
email    : Nasy <nasyxx+python@gmail.com>
filename : types.py
project  : nacf
license  : GPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
# Standard Library
from typing import (IO, Dict, List, Tuple, Union, TypeVar,
                    Callable, Iterable, Optional, ByteString,)

# Web Packages
from requests_html import HTML
from requests_html import Element as Ele
from requests_html import HTMLResponse as Res

# Normal
a = TypeVar("a")
b = TypeVar("b")

# function
RT = TypeVar("RT")
Func = Callable[..., RT]

Reses = Iterable[Res]
RFunc = Callable[..., Res]
RsFunc = Callable[..., Reses]

Eles = Iterable[Ele]
EFunc = Callable[..., Union[Ele, Eles]]

SFunc = Callable[..., Union[str, Iterable[str]]]

# request
Data = Optional[Union[Dict[str, a], ByteString, IO]]
Json = Dict[str, a]
Url = str
Urls = Iterable[Url]

__all__ = [
    "Res",
    "Reses",
    "HTML",
    "a",
    "b",
    "RT",
    "Func",
    "RFunc",
    "RsFunc",
    "SFunc",
    "Data",
    "Json",
    "Url",
    "Urls",
    "List",
    "Tuple",
]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['nacf']

package_data = \
{'': ['*']}

install_requires = \
['nalude>=0.2.0,<0.3.0', 'requests_html>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'nacf',
    'version': '1.0.1',
    'description': 'Nasy Crawler Framework -- Never had such a pure crawler.',
    'long_description': '# Table of Contents\n\n-   [Prologue](#org6dc9402)\n-   [Packages](#org37032fa)\n-   [Development Process](#org1eb07f9)\n    -   [Http Functions](#orgd862ea6)\n        -   [Get](#orgb88fa50)\n        -   [Post](#orgd091da7)\n        -   [Bugs](#orga99911a)\n            -   [Fix an error from inspect.Parameter which caused the function parallel down.](#orga7af803):err:1:\n    -   [Docs](#orga5017d7)\n        -   [Usage](#org21efc61)\n-   [Epoligue](#org1e897d7)\n    -   [History](#orgba8919d)\n        -   [Version 1.0.0](#orgb21403b)\n        -   [Version 0.1.2](#orgcd42135)\n        -   [Version 0.1.1](#orgce6ff7d)\n        -   [Version 0.1.0](#orgcc23226)\n\n\n\n<a id="org6dc9402"></a>\n\n# Prologue\n\nNever had such a pure crawler like this `nacf`.\n\nAlthough I often write crawlers, I don&rsquo;t like to use huge frameworks, such as scrapy, but prefer\nsimple `requests+bs4` or more general `requests_html`.  However, these two are inconvenient for a\ncrawler.  E.g. Places, such as error retrying or parallel crawling, need to be handwritten by\nmyself.  It is not very difficult to write it while writing too much can be tedious.  Hence I\nstarted writing this nacf (Nasy Crawler Framework), hoping to simplify some error retrying or\nparallel writing of crawlers.\n\n\n<a id="org37032fa"></a>\n\n# Packages\n\n<table>\n<caption class="t-above"><span class="table-number">Table 1:</span> Packages</caption>\n\n<colgroup>\n<col  class="org-left">\n\n<col  class="org-right">\n\n<col  class="org-left">\n</colgroup>\n<thead>\n<tr>\n<th scope="col" class="org-left">Package</th>\n<th scope="col" class="org-right">Version</th>\n<th scope="col" class="org-left">Description</th>\n</tr>\n</thead>\n\n<tbody>\n<tr>\n<td class="org-left">requests-html</td>\n<td class="org-right">0.9.0</td>\n<td class="org-left">HTML Parsing for Humans.</td>\n</tr>\n\n\n<tr>\n<td class="org-left">nalude</td>\n<td class="org-right">0.2.0</td>\n<td class="org-left">A standard module.  Inspired by Haskell&rsquo;s Prelude.</td>\n</tr>\n</tbody>\n</table>\n\n\n<a id="org1eb07f9"></a>\n\n# Development Process\n\n\n<a id="orgd862ea6"></a>\n\n## DONE Http Functions\n\n<p><span class="timestamp-wrapper"><span class="timestamp-kwd">CLOSED:</span> <span class="timestamp">&lt;Thu Feb 28 20:51:00 2019&gt;</span></span></p>\n\n\n<a id="orgb88fa50"></a>\n\n### DONE Get\n\n<p><span class="timestamp-wrapper"><span class="timestamp-kwd">CLOSED:</span> <span class="timestamp">&lt;Tue Dec 25 17:36:00 2018&gt;</span></span></p>\n\n\n<a id="orgd091da7"></a>\n\n### DONE Post\n\n<p><span class="timestamp-wrapper"><span class="timestamp-kwd">CLOSED:</span> <span class="timestamp">&lt;Thu Feb 28 20:44:00 2019&gt;</span></span></p>\n\n\n<a id="orga99911a"></a>\n\n### DONE Bugs\n\n<p><span class="timestamp-wrapper"><span class="timestamp-kwd">CLOSED:</span> <span class="timestamp">&lt;Thu Feb 28 20:51:00 2019&gt;</span></span></p>\n\n\n<a id="orga7af803"></a>\n\n#### DONE Fix an error from inspect.Parameter which caused the function parallel down.     :err:1:\n\n<p><span class="timestamp-wrapper"><span class="timestamp-kwd">CLOSED:</span> <span class="timestamp">&lt;Wed Dec 26 20:26:00 2018&gt;</span></span></p>\n\n\n<a id="orga5017d7"></a>\n\n## NEXT Docs\n\n\n<a id="org21efc61"></a>\n\n### NEXT Usage\n\n\n<a id="org1e897d7"></a>\n\n# Epoligue\n\n\n<a id="orgba8919d"></a>\n\n## History\n\n\n<a id="orgb21403b"></a>\n\n### Version 1.0.0\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Thu Feb 28, 2019&gt;</span></span>\n-   **Changes:** Now, old HTTP methods (`get` and `post`) cannot accept multiple URLs. Instead, we can use `gets` and `posts`.\n-   **Adds:** -   `nacf.html`\n    -   `nacf.json`\n    -   `nacf.gets`\n    -   `nacf.posts`\n-   **Includes:** -   `nalude`\n\n\n<a id="orgcd42135"></a>\n\n### Version 0.1.2\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Wed Dec 26, 2018&gt;</span></span>\n-   **Fixed:** `inspect.Parameter` error in last version.\n\n\n<a id="orgce6ff7d"></a>\n\n### Version 0.1.1\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Wed Dec 26, 2018&gt;</span></span>\n-   **Ignored:** An error caused by `inspect.Parameter`\n-   **Help Wanted:** Can someone help me about the Parameter?\n\n\n<a id="orgcc23226"></a>\n\n### Version 0.1.0\n\n-   **Date:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Sun Dec 23, 2018&gt;</span></span>\n-   **Commemorate Version:** First Version\n    -   Basic Functions.\n',
    'author': 'Nasy',
    'author_email': 'nasyxx+nacf@gmail.com',
    'url': 'https://github.com/nasyxx/nacf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec 24, 2018
update   : Feb 28, 2019
email    : Nasy <nasyxx+python@gmail.com>
filename : models.py
project  : nacf
license  : GPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
# Standard Library
from inspect import Parameter
from functools import wraps
from itertools import chain, cycle
from multiprocessing.dummy import Pool

# Web Packages
from requests_html import HTMLSession as s

# Local Packages
from .tools import format_url, generate_sig
from .types import (RT, HTML, Ele, Res, Url, Data, Eles, Func, Json,
                    List, Urls, EFunc, Reses, RFunc, SFunc, Tuple,
                    Union, RsFunc, Callable, Iterable, Optional, a, b,)


def urls(urls_: Urls, **kwds: a) -> Callable[[Func], Func]:
    """Feed urls to next function."""

    def decorator(func: Func) -> Func:
        """Decorate function func."""

        @wraps(func)
        def wrapper() -> RT:
            """Wrap function func."""
            return func(_urls=urls_, **kwds)

        return wrapper

    return decorator


def parallel(n: Optional[int] = None, **kwds: a) -> Callable[[Func], Func]:
    """Run parallelly."""

    def decorator(func: Func) -> Func:
        """Decorate function func."""

        @wraps(func)
        def wrapper(
            _urls: Urls,
            *,
            _datas: Optional[Iterable[Data]] = None,
            _jsons: Optional[Iterable[Json]] = None,
            **_kwds: b,
        ) -> RT:
            """Wrap function func."""

            def runner(udj: Tuple[Url, Optional[Data], Optional[Json]]) -> RT:
                """Run func wrapper."""
                url, data, json = udj
                return (
                    func(url, _data=data, _json=json, **_kwds, **kwds)
                    if data or json
                    else func(url, **_kwds, **kwds)
                )

            with Pool(n) as pool:
                res = pool.map(
                    runner,
                    zip(
                        _urls,
                        _datas or cycle([None]),
                        _jsons or cycle([dict()]),
                    ),
                )
            return res

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_datas",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[Iterable[Data]]",
                ),
                Parameter(
                    "_jsons",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[Iterable[Json]]",
                ),
            ),
        )

        return wrapper

    return decorator


def get(url: Url = None, **kwds: a) -> Callable[[RFunc], RFunc]:
    """Get url."""

    def decorator(func: RFunc) -> RFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(
            _url: Optional[Url] = None, *, _s: Optional[s] = None
        ) -> Res:
            """Wrap function func."""
            return func((_s or s()).get(format_url(_url or url), **kwds))

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_url",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Url]",
                ),
                Parameter(
                    "_s",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[HTMLSession]",
                ),
            ),
        )

        return wrapper

    return decorator


def gets(urls: Urls = None, **kwds: a) -> Callable[[RsFunc], RsFunc]:
    """Get url."""

    def decorator(func: RsFunc) -> RsFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(
            _urls: Optional[Urls] = None, *, _s: Optional[s] = None
        ) -> Reses:
            """Wrap function func."""
            return func(
                map(
                    lambda u: (_s or s()).get(format_url(u)),
                    _urls or urls or [""],
                ),
                **kwds,
            )

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_urls",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Urls]",
                ),
                Parameter(
                    "_s",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[HTMLSession]",
                ),
            ),
        )

        return wrapper

    return decorator


def post(
    url: Optional[Url] = None,
    *,
    data: Data = None,
    json: Json = None,
    **kwds: a,
) -> Callable[[RFunc], RFunc]:
    """Post url with data and json."""

    def decorator(func: RFunc) -> RFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(
            _url: Optional[Url] = None,
            _data: Optional[Data] = None,
            _json: Optional[Json] = None,
            *,
            _s: Optional[s] = None,
        ) -> Res:
            """Wrap function func."""
            return func(
                (_s or s()).post(format_url(url), data=data, json=json, **kwds)
            )

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_url",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Url]",
                ),
                Parameter(
                    "_data",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Data]",
                ),
                Parameter(
                    "_json",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Json]",
                ),
                Parameter(
                    "_s",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[HTMLSession]",
                ),
            ),
        )

        return wrapper

    return decorator


def posts(
    urls: Optional[Urls] = None,
    *,
    datas: Optional[Iterable[Data]] = None,
    jsons: Optional[Iterable[Json]] = None,
    **kwds: a,
) -> Callable[[RsFunc], RsFunc]:
    """Post url with data and json."""

    def decorator(func: RsFunc) -> RsFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(
            _urls: Optional[Urls] = None,
            *,
            _datas: Optional[Iterable[Data]] = None,
            _jsons: Optional[Iterable[Json]] = None,
            _s: Optional[s] = None,
        ) -> Res:
            """Wrap function func."""

            def _post(udj: Tuple[Url, Data, Json]) -> Res:
                """Make a post."""
                url, data, json = udj
                return (_s or s()).post(
                    format_url(url), data=data, json=json, **kwds
                )

            return func(
                map(
                    _post,
                    zip(
                        _urls or urls or [""],
                        _datas or datas or cycle([None]),
                        _jsons or jsons or cycle([dict()]),
                    ),
                )
            )

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_urls",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Urls]",
                ),
                Parameter(
                    "_datas",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Iterable[Data]]",
                ),
                Parameter(
                    "_jsons",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Optional[Iterable[Json]]",
                ),
                Parameter(
                    "_s",
                    Parameter.KEYWORD_ONLY,
                    default=None,
                    annotation="Optional[HTMLSession]",
                ),
            ),
        )

        return wrapper

    return decorator


def css(
    selector: str = "*",
    *,
    containing: Union[str, List[str]] = None,
    clean: bool = False,
    first: bool = False,
    _encoding: str = None,
) -> Callable[[EFunc], EFunc]:
    """Css response."""

    def decorator(func: EFunc) -> EFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(_res: Union[Res, Reses]) -> Union[Ele, Eles]:
            """Wrap function func."""
            return (
                func(
                    _res.html.find(
                        selector,
                        containing=containing,
                        clean=clean,
                        first=first,
                        _encoding=_encoding,
                    )
                )
                if isinstance(_res, Res)
                else func(
                    (lambda ele: ele if first else chain.from_iterable(ele))(
                        map(
                            lambda r: r.html.find(
                                selector,
                                containing=containing,
                                clean=clean,
                                first=first,
                                _encoding=_encoding,
                            ),
                            _res,
                        )
                    )
                )
            )

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_res",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Res",
                ),
            ),
        )

        return wrapper

    return decorator


def xpath(
    selector: str,
    *,
    clean: bool = False,
    first: bool = False,
    _encoding: str = None,
) -> Callable[[EFunc], EFunc]:
    """Css response."""

    def decorator(func: EFunc) -> EFunc:
        """Decorate function func."""

        @wraps(func)
        def wrapper(_res: Union[Res, Reses]) -> Union[Ele, Eles]:
            """Wrap function func."""
            return func(
                _res.html.xpath(
                    selector, clean=clean, first=first, _encoding=_encoding
                )
                if isinstance(_res, Res)
                else (lambda ele: ele if first else chain.from_iterable(ele))(
                    map(
                        lambda r: r.html.xpath(
                            selector,
                            clean=clean,
                            first=first,
                            _encoding=_encoding,
                        ),
                        _res,
                    )
                )
            )

        setattr(
            wrapper,
            "__signature__",
            generate_sig(
                func,
                Parameter(
                    "_res",
                    Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation="Res",
                ),
            ),
        )

        return wrapper

    return decorator


def html(func: SFunc) -> SFunc:
    """Get html from response."""

    @wraps(func)
    def wrapper(_res: Union[Res, Reses]) -> Union[HTML, Iterable[HTML]]:
        """Wrap function func."""
        return (
            func(_res.html)
            if isinstance(_res, Res)
            else map(lambda r: r.html, _res)
        )

    setattr(
        wrapper,
        "__signature__",
        generate_sig(
            func,
            Parameter(
                "_res",
                Parameter.POSITIONAL_OR_KEYWORD,
                default=None,
                annotation="Res",
            ),
        ),
    )

    return wrapper


def json(
    func: Callable[..., Union[Json, Iterable[Json]]]
) -> Callable[[Union[Res, Reses]], Union[Json, Iterable[Json]]]:
    """Get json from response."""

    @wraps(func)
    def wrapper(_res: Union[Res, Reses]) -> Union[Json, Iterable[Json]]:
        """Wrap function func."""
        return (
            func(_res.json())
            if isinstance(_res, Res)
            else map(lambda r: r.json(), _res)
        )

    setattr(
        wrapper,
        "__signature__",
        generate_sig(
            func,
            Parameter(
                "_res",
                Parameter.POSITIONAL_OR_KEYWORD,
                default=None,
                annotation="Res",
            ),
        ),
    )

    return wrapper


def text(func: SFunc) -> SFunc:
    """Get text from response or element(s) from css/xpath."""

    @wraps(func)
    def wrapper(_ele: Union[Res, Ele, Eles]) -> Union[str, Iterable[str]]:
        """Wrap function func."""
        return func(
            _ele.text
            if isinstance(_ele, Ele) or isinstance(_ele, Res)
            else map(lambda e: e.text, _ele)
        )

    setattr(
        wrapper,
        "__signature__",
        generate_sig(
            func,
            Parameter(
                "_ele",
                Parameter.POSITIONAL_OR_KEYWORD,
                default=None,
                annotation="Union[Res, Ele, Eles]",
            ),
        ),
    )

    return wrapper
